var AspectScript = (function(){

    // ---- PRIVATE STATIC ----

    // Tracing
    var tracerOn = new (function(){
        this.out = "";

        this.jp = function(jp){
            //print(jp)
            this.out += "debugInfo.push({k:'jp',level:" + jp.metaLevel + ",toString:'" + jp.toString() + "'});";
        };

        this.pcFalseByReentrancy = function(aspect, jp){
            this.out += "debugInfo.push({k:'pc',id:" + aspect.id + ",level:" + (jp.metaLevel + 1) + ",returned:false,reason:'reentrance'});";
        };

        this.pcEval = function(aspect, jp, env){
            var ret = (env ? true : false);
            var reason = (env ? 'env' : 'bool');
            var level = (jp.metaLevel + 1);
            this.out += "debugInfo.push({k:'pc',id:" + aspect.id + ",level:" + level + ",returned:" + ret + ",reason:'" + reason + "'});";
        };

        this.advEval = function(aspect, jp){
            var level = (jp.metaLevel + 1);
            this.out += "debugInfo.push({k:'adv',id:" + aspect.id + ",level:" + level + "});";
        };

        this.asps = function(asps, newasps){
            var o = asps.map(function(asp){
                return asp.id;
            });
            var n = newasps.map(function(asp){
                return asp.id;
            });
            this.out += "debugInfo.push({k:'aspects',old:[" + o + "],neu:[" + n + "]});";
        };

        this.proceed = function(jp){
            this.out += "debugInfo.push({k:'proceed',level:" + jp.metaLevel + ",toString:'" + jp.toString() + "'});";
        };

        this.deploy = function(level){
            this.out += "debugInfo.push({k:'deploy', level:" + level + ",type:'GLOBAL'});";
        };

        this.enterSS_C = function(jp, asps, newasps){
            var level = jp.metaLevel;
            var o = asps.map(function(asp){
                return asp.id;
            });
            var n = newasps.map(function(asp){
                return asp.id;
            });
            this.out += "debugInfo.push({k:'enterss_c', level:" + level + ",old:" + o + ",neu:" + n + "});";
        };

        this.chainPush = function(chain, elem){
            var o = chain.map(function(elem){
                return elem[0].id;
            });
            this.out += "debugInfo.push({k:'chain_push',aspects:[" + o + "],aspect:" + elem[0].id + ",jp:'" + elem[1] + "'});";
        };

        this.chainPop = function(chain){
            var o = chain.map(function(elem){
                return elem[0].id;
            });
            this.out += "debugInfo.push({k:'chain_pop',aspects:[" + o + "]});";
        };

        this.pushCtx = function(obj, fun){
            var o = obj.map(function(elem){
                return elem.id;
            });
            var f = fun.map(function(elem){
                return elem.id;
            });
            this.out += "debugInfo.push({k:'ctx_push',obj:[" + o + "],fun:[" + f + "]});";
        };

        this.popCtx = function(obj, fun){
            var o = obj.map(function(elem){
                return elem.id;
            });
            var f = fun.map(function(elem){
                return elem.id;
            });
            this.out += "debugInfo.push({k:'ctx_pop',obj:[" + o + "],fun:[" + f + "]});";
        };

        this.clear = function(){
            this.out = "";
        };

        this.dump = function(){
            print("var debugInfo = []; " + this.out);
        };

        this.enable = function(){
        };

        this.disable = function(){
            tracer = AspectScript.tracer = tracerOff;
        };
    })();
    var tracerOff = new (function(){
        var emptyFun = function(){
        };

        for(var i in tracerOn){
            if(tracerOn.hasOwnProperty(i)){
                this[i] = emptyFun;
            }
        }

        this.clear = function(){
            tracerOn.dump();
        };

        this.dump = function(){
            tracerOn.dump();
        };

        this.enable = function(){
            tracer = AspectScript.tracer = tracerOn;
        };

        this.disable = emptyFun;
    })();
    var tracer = tracerOff;

    // JoinPoint kind constants
    var JP_NEW = 0;
    var JP_INIT = 1;
    var JP_CALL = 2;
    var JP_EXEC = 3;
    var JP_PROP_READ = 4;
    var JP_PROP_WRITE = 5;
    var JP_VAR_READ = 6;
    var JP_VAR_WRITE = 7;
    var JP_CUSTOM = 8;

    //global aspect environment
    var aspects = new AE();

    function AE(){
        var aspects = [];

        this.get = function(level){
            growAE(level);
            return aspects[level];
        };

        this.add = function(level, asp){
            growAE(level);
            aspects[level].push(asp);
        };

        this.remove = function(level, asp){
            growAE(level);
            var aspectsInLevel = aspects[level];
            for(var i = 0, n = aspectsInLevel.length; i < n; ++i){
                if(aspectsInLevel[i] === asp){
                    aspectsInLevel.splice(i, 1);
                    break;
                }
            }
        };

        this.set = function(level, asps){
            growAE(level);
            aspects[level] = asps;
        };

        this.isEmpty = function(level){
            growAE(level);
            return aspects[level].length == 0;
        };

        function growAE(level){
            for(var i = aspects.length; i <= level; ++i){
                aspects[i] = [];
            }
        }
    }

    //to allow the construction of the parent link
    var currentJoinPoint = null;

    //meta (indicate the level the application is currently executing in)
    var currentMetaLevel = 0;

    //to optimize reentrance checking in isReentering
    var aspectsChain = [];
    aspectsChain.push = function(elem){
        tracer.chainPush(this, elem);
        Array.prototype.push.call(this, elem);
    };
    aspectsChain.pop = function(){
        tracer.chainPop(this);
        Array.prototype.pop.call(this);
    };

    //keys
    var SILENCE_NEXT_EXEC_KEY = {};

    //util
    function returnTrueFunction(){
        return true;
    }

    function removeFromArray(array, elem){
        for(var i = 0; i < array.length; ++i){
            if(array[i] === elem){
                array.splice(i, 1);
                return;
            }
        }
    }

    var EMPTY_ARRAY = [];  //do not return it to the user

    var weavingEnabled = false;

    function Constructor(){

        this.BEFORE = -1;
        this.AROUND = 0;
        this.AFTER = 1;

        this.globalObject = (function(){
            return this;
        })();

        // ---- PRIVATE  ----
        this.tracer = tracer; //this is necessary to give external access to the tracer

        // Weaving

        function weave(jp){
            tracer.jp(jp);
            currentJoinPoint = jp;

            try{
                enterExecContext();
                var union = u();

                var ssEnabled = aspectsWithSS && 1 == 1; //enable SS only when it is necessary
                if(ssEnabled){ //todo: strategy pattern
                    enterSS();
                }
                enterInitContext();

                //tracer.asps(asd, []);
                var areThereAspects = weavingEnabled && union.length > 0;
                if(!areThereAspects){
                    tracer.proceed(jp);
                    return jp._proceed();
                }

                var finalAspects = match(currentJoinPoint, union);
                if(finalAspects.length == 0){
                    return currentJoinPoint._proceed();
                }
                else{
                    return currentJoinPoint.chain(finalAspects);
                }
            }
            finally{
                if(ssEnabled){
                    exitSS();
                }

                exitContext();
                currentJoinPoint = jpPool.release(currentJoinPoint).parentAllLevel;
            }
        }

        function match(jp, aspects){
            var matchresult = [];

            //todo: maybe use only one "up"?
            for(var i = 0, n = aspects.length; i < n; ++i){
                var aspect = aspects[i];
                //f
                var f = AspectScript.up(function up_f(){
                    return aspect.f === true || (aspect.f !== false && aspect.f(jp, AspectScript.emptyEnv));
                });
                if(!f){
                    continue;
                }
                //pc
                var env = AspectScript.up(function up_executePointcut(){
                    return executePointcut(aspect, jp, AspectScript.emptyEnv);
                });
                if(env){
                    matchresult.push([aspect, env]);
                }
            }

            return matchresult;
        }

        function executePointcut(aspect, jp, env){
            if(isReentering(jp, aspect)){
                tracer.pcFalseByReentrancy(aspect, jp);
                return false;
            }

            aspectsChain.push([aspect, jp]);
            var newEnv;
            try{
                newEnv = aspect.pointcut(jp, env); //the actual context is determined by the functions created in Aspect!
            }
            finally{
                aspectsChain.pop();
            }

            tracer.pcEval(aspect, jp, newEnv);

            if(newEnv && (newEnv instanceof Env)){
                return newEnv;
            }

            //return-value normalization: env or boolean
            //if the value is "only" true, we preserve the previous environment
            if(newEnv){
                if(env){
                    return env;
                }
                else{
                    throw "lost env";
                }
            }
            else{
                return false;
            }
        }

        // Union
        function u(){
            var globalAE = aspects.get(currentJoinPoint.metaLevel);
            var objAspects = ddAspects.getAspectsForObjInContext();
            var funAspects = ddAspects.getAspectsForFunInContext();

            if(globalAE.length + objAspects.length + funAspects.length == 0){
                return [];
            }
            //concat all
            var currentAE = merge3(globalAE, funAspects, objAspects);

            //union
            var uresult = [];

            var idx = 0;
            for(var i = 0, n = currentAE.length; i < n; ++i){
                if(idx == 0 || uresult[idx - 1].id !== currentAE[i].id){
                    uresult[idx++] = currentAE[i];
                }
            }

            return uresult;
        }

        function merge2(a, b){
            var res = [];

            var na = a.length;
            var nb = b.length;
            var x = 0, y = 0;
            while(x < na && y < nb){
                if(a[x].id < b[y].d){
                    res.push(a[x++]);
                }
                else{
                    res.push(b[y++]);
                }
            }

            while(x < na){
                res.push(a[x++]);
            }
            while(y < nb){
                res.push(b[y++]);
            }

            return res;
        }

        function merge3(a, b, c){
            var res = [];

            var na = a.length;
            var nb = b.length;
            var nc = c.length;
            var x = 0, y = 0, z = 0;
            //merge 3
            while(x < na && y < nb && z < nc){
                res.push((a[x].id < b[y].id) ?
                         (a[x].id < c[z].id ? a[x++] : c[z++]) :
                         (b[y].id < c[z].id ? b[y++] : c[z++]));
            }

            //merge 2
            while(x < na && y < nb){
                res.push((a[x].id < b[y].id) ? a[x++] : b[y++]);
            }
            while(x < na && z < nc){
                res.push((a[x].id < c[z].id) ? a[x++] : c[z++]);
            }
            while(y < nb && z < nc){
                res.push((b[y].id < c[z].id) ? b[y++] : c[z++]);
            }

            //final copy
            while(x < na){
                res.push(a[x++]);
            }
            while(y < nb){
                res.push(b[y++]);
            }
            while(z < nc){
                res.push(c[z++]);
            }

            return res;
        }

        // Current execution context
        // (init and exec are splitted because of their interaction with enterSS: INIT!!)
        function enterExecContext(){
            if(currentJoinPoint.kind == JP_EXEC){
                ddAspects.pushObjectsInContext(currentJoinPoint.target, currentJoinPoint.fun);
            }
        }

        function enterInitContext(){
            if(currentJoinPoint.kind == JP_INIT){
                ddAspects.pushObjectsInContext(currentJoinPoint.target, currentJoinPoint.fun);
            }
        }

        function exitContext(){
            if(currentJoinPoint.kind == JP_EXEC || currentJoinPoint.kind == JP_INIT){
                ddAspects.popObjectsInContext();
            }
        }

        // Scoping Strategies
        var oldGlobalAspects = [];

        function enterSS(){
            if(!(currentJoinPoint.kind == JP_CALL || currentJoinPoint.kind == JP_INIT)){
                return;
            }

            var dasps = u();
            var newDasps = [], i;

            //c
            if(currentJoinPoint.kind == JP_CALL){
                AspectScript.up(function up_c(){
                    for(i = 0; i < dasps.length; ++i){
                        if(dasps[i].c === true || (dasps[i].c !== false && dasps[i].c(currentJoinPoint, AspectScript.emptyEnv))){
                            newDasps.push(dasps[i]);
                        }
                    }
                });

                oldGlobalAspects.push(aspects.get(currentJoinPoint.metaLevel)); //remember global aspects
                aspects.set(currentJoinPoint.metaLevel, newDasps); //and update them
                tracer.enterSS_C(currentJoinPoint, dasps, newDasps);
            }
            //d
            else{ //if(currentJoinPoint.isInit){
                AspectScript.up(function up_d(){
                    for(i = 0; i < dasps.length; ++i){
                        var jp = currentJoinPoint.parent; //TODO
                        if(dasps[i].d === true || (dasps[i].d !== false && dasps[i].d(jp, AspectScript.emptyEnv))){
                            newDasps.push(dasps[i]);
                        }
                    }
                });

                //deploy aspects on the object
                if(newDasps.length != 0){
                    ddAspects.put(currentJoinPoint.target, newDasps);
                }
            }
        }

        function exitSS(){
            if(currentJoinPoint.kind == JP_CALL){
                var i;
                //old dynamic aspects in global array
                var dynamic = oldGlobalAspects.pop();
                for(i = 0; i < dynamic.length; ++i){
                    if(dynamic[i].isGlobal){
                        dynamic.splice(i, 1);
                    }
                }
                //new global aspects
                var global = aspects.get(currentJoinPoint.metaLevel);
                for(i = 0; i < global.length; ++i){
                    if(!global[i].isGlobal){
                        global.splice(i, 1);
                    }
                }

                aspects.set(currentJoinPoint.metaLevel, merge2(global, dynamic));
            }
        }

        // Environments
        this.emptyEnv = new Env([]);

        function Env(bindings){

            this.bind = function(){
                var b = {};
                // case: bind(id,value)
                if(arguments.length == 2 && typeof(arguments[0]) != Array){
                    b[arguments[0]] = arguments[1];
                }
                else{ //case:bind([id1,value1],[id2,value2],..)
                    for(var i = 0; i < arguments.length; ++i){
                        var arg = arguments[i];
                        b[arg[0]] = arg[1];
                    }
                }

                for(var name in bindings){
                    if(!(name in b)){
                        b[name] = bindings[name];
                    }
                }
                return new Env(b);
            };

            function createBind(name){
                return function (){
                    return bindings[name];
                };
            }

            //define getters to support env.id
            for(var name in bindings){
                this.__defineGetter__(name, createBind(name));
            }

            this.getAll = function(){
                var b = [];

                for(var name in bindings){
                    b.push([name,bindings[name]]);
                }

                return b;
            };

            this.get = function(name){
                for(var n in bindings){
                    if(n === name){
                        return bindings[n];
                    }
                }

                throw "binding not found: " + name;
            };

            this.unbind = function(){
                var args = Array.prototype.slice.call(arguments);

                var b = {};
                for(var name in bindings){
                    if(args.indexOf(name) < 0){
                        b[name] = bindings[name];
                    }
                }

                return new Env(b);
            };

            this.toString = function(){
                var str = "";
                for(var n in bindings){
                    str += n + "=" + bindings[n] + "\n";
                }
                if(str === "")
                    str += "[empty]";

                return str;
            };
        }

        // Join Points
        var jpPool = new JPPool();

        //disable this if it is necessary to save references to past jps
        var ENABLE_JP_POOL = true;

        function JPPool(){
            var constructors = [];
            constructors[JP_NEW] = function (kind, thunk, constructor, args, parent){
                return new NewJoinPoint(kind, thunk, constructor, args, parent);
            };
            constructors[JP_INIT] = function (kind, target, args, wrapper, constructor, parent){
                return new InitJoinPoint(kind, target, args, wrapper, constructor, parent);
            };
            constructors[JP_CALL] = function (kind, target, fun, args, context, parent){
                return new CallJoinPoint(kind, target, fun, args, context, parent);
            };
            constructors[JP_EXEC] = function (kind, target, args, wrapperFun, originalFun, parent){
                return new ExecJoinPoint(kind, target, args, wrapperFun, originalFun, parent);
            };
            constructors[JP_PROP_READ] = function (kind, target, name, parent){
                return new PropertyReadJoinPoint(kind, target, name, parent);
            };
            constructors[JP_PROP_WRITE] = function (kind, target, name, value, parent){
                return new PropertyWriteJoinPoint(kind, target, name, value, parent);
            };
            constructors[JP_VAR_READ] = function (kind, name, value, parent){
                return new VarReadJoinPoint(kind, name, value, parent);
            };
            constructors[JP_VAR_WRITE] = function (kind, name, value, oldValue, parent){
                return new VarWriteJoinPoint(kind, name, value, oldValue, parent);
            };
            constructors[JP_CUSTOM] = function (kind, name, args, proceed, parent){
                return new CustomJoinPoint(kind, name, args, proceed, parent);
            };

            var pool = [];

            for(var i = 0; i < constructors.length; ++i){
                pool[i] = [];

                //check whether the optimization in "get" will work
                if(constructors[i].length > 6){
                    throw "'get' optimization will not work";
                }
            }

            this.get = function(kind){
                var jp;
                if(ENABLE_JP_POOL && pool[kind].length != 0){
                    jp = pool[kind].pop();
                    //this is TWICE as fast as "inner.reinit.apply(inner, arguments);"
                    //max 6 arguments (see "for" above)
                    jp.reinit(
                            arguments[0], arguments[1], arguments[2],
                            arguments[3], arguments[4], arguments[5]
                            );
                }
                else{
                    jp = constructors[kind].apply(null, arguments);
                }

                return jp;
            };

            this.release = function(jp){
                pool[jp.kind].push(jp);
                return jp;
            };
        }

        function JoinPoint(){

            this._reinit = function(_kind, _parent){
                this.kind = _kind;
                this.parentAllLevel = _parent;
                this.metaLevel = currentMetaLevel + 1;
                this.finalResult = null;
                //chain
                this.adviceIdx = 0;
                this.advices = null;
            };

            this.chain = function(advices){
                this.adviceIdx = 0;
                this.advices = advices;
                var jp = this;

                return AspectScript.up(function adv_up_around(){
                    return jp.proceed();
                });
            };

            var originalCtx = null;

            this.proceed = function(){
                var jp = this;

                //no advices => original computation (without down)
                if(this.advices === null){
                    //tracer.proceed(jp);
                    return jp._proceed.apply(jp, args); //jp._proceed(arg1, arg2,...)
                }

                //no more advices => original computation
                if(this.adviceIdx == this.advices.length){
                    var args = arguments;
                    return AspectScript.down(function(){
                        ddAspects.pushObjectsInContext(originalCtx[0], originalCtx[1]); //restore original context
                        var previousPair = (aspectsChain.length > 0) ? aspectsChain.pop() : null; //todo
                        try{
                            tracer.proceed(jp);
                            return jp._proceed.apply(jp, args); //jp._proceed(arg1, arg2,...)
                        }
                        finally{
                            ddAspects.popObjectsInContext(); //restore original context
                            if(previousPair != null){
                                aspectsChain.push(previousPair);
                            }
                        }
                    });
                }

                if(this.adviceIdx == 0){ //remember original context (for last proceed)
                    originalCtx = ddAspects.peekObjectsInContext();
                }

                //advice execution
                var advIdx = this.advices.length - this.adviceIdx - 1;
                var aspect = this.advices[advIdx][0];
                var env = this.advices[advIdx][1];
                ++this.adviceIdx;

                var previousPair = (this.adviceIdx > 1) ? aspectsChain.pop() : null;
                aspectsChain.push([aspect,jp]);
                try{
                    tracer.advEval(aspect, jp);
                    return aspect.advice(jp, env); //the actual context is determined by the functions created in Aspect!
                }
                finally{
                    aspectsChain.pop();
                    if(previousPair != null){
                        aspectsChain.push(previousPair);
                    }
                    --this.adviceIdx;
                }
            };

            this.__defineGetter__("parent", function(){
                var parent = this.parentAllLevel;
                while(parent !== null && parent.metaLevel !== this.metaLevel){
                    parent = parent.parentAllLevel;
                }
                return parent;
            });

            var getters = {
                New:JP_NEW, Init:JP_INIT,
                Call:JP_CALL, Exec:JP_EXEC,
                PropRead: JP_PROP_READ, PropWrite:JP_PROP_WRITE,
                VarRead: JP_VAR_READ, VarWrite:JP_VAR_WRITE
            };

            for(var k in getters){
                this["is" + k] = createFunctionKind(this, getters[k]);
            }

            this.isCustom = function(name){
                return this.kind == JP_CUSTOM && this.name === name;
            };

            this.copyFrom = function(jp){
                this.metaLevel = jp.metaLevel;
                this.finalResult = jp.finalResult;
                this.adviceIdx = jp.adviceIdx;
                this.advices = jp.advices;
            };

            this.toString = function(){
                var strs = [];
                strs[JP_NEW] = "new";
                strs[JP_INIT] = "init";
                strs[JP_CALL] = "call";
                strs[JP_EXEC] = "exec";
                strs[JP_VAR_READ] = "vr";
                strs[JP_VAR_WRITE] = "vw";
                strs[JP_PROP_READ] = "pr";
                strs[JP_PROP_WRITE] = "pw";
                strs[JP_CUSTOM] = "custom";
                return "[" + strs[this.kind] + "]";
            };
        }

        function createFunctionKind(jp, value){
            return function(){
                return jp.kind == value;
            };
        }

        function NewJoinPoint(_kind, _thunk, _fun, _args, _parent){
            JoinPoint.call(this);

            //private
            var thunk;
            var args;

            this.reinit = function(_kind, _thunk, _fun, _args, _parent){
                this._reinit(_kind, _parent);
                thunk = _thunk;
                this.fun = _fun;
                args = _args;
            };

            this.reinit(_kind, _thunk, _fun, _args, _parent);

            this.__defineGetter__("args", function(){
                return Array.prototype.slice.call(args);
            });

            this._proceed = function(){
                var actualArgs = (arguments.length == 0) ? args : arguments;
                return thunk.apply(null, [this.fun].concat(actualArgs)); //call the thunk
            };

            this.clone = function(){
                var r = jpPool.get(this.kind, thunk, this.fun, args, this.parentAllLevel);
                r.copyFrom(this);
                return r;
            };
        }

        function InitJoinPoint(_kind, _target, _args, _wrapper, _constructor, _parent){
            JoinPoint.call(this);

            //private
            var args;
            var fun;

            this.reinit = function(_kind, _target, _args, _wrapper, _constructor, _parent){
                this._reinit(_kind, _parent);
                this.target = _target;
                args = _args;
                fun = _constructor;
                this.fun = _wrapper;
            };

            this.reinit(_kind, _target, _args, _wrapper, _constructor, _parent);

            this.__defineGetter__("args", function(){
                return Array.prototype.slice.call(args);
            });

            this._proceed = function(){
                var actualArgs = (arguments.length == 0) ? args : arguments;
                return fun.apply(this.target, actualArgs);
            };

            this.clone = function(){
                var r = jpPool.get(this.kind, this.target, args, this.fun, fun, this.parentAllLevel);
                r.copyFrom(this);
                return r;
            };
        }

        function CallJoinPoint(_kind, _target, _fun, _args, _context, _parent){
            JoinPoint.call(this);

            //private
            var args;
            var methods;

            this.reinit = function(_kind, _target, _fun, _args, _context, _parent){
                this._reinit(_kind, _parent);
                this.target = _target;
                this.fun = _fun;
                this.context = _context;
                args = _args;
                methods = null;
                this.reflective = false;

                //reflective?
                if(this.target instanceof Function){
                    //fun.call(obj, arg1, arg2...)
                    if(this.fun === Function.prototype.call){
                        this.fun = this.target;
                        this.target = args[0] != null ? args[0] : AspectScript.globalObject;  //target is args[0]
                        args.splice(0, 1); //remove target from the args
                        this.reflective = true;
                    }
                    //fun.apply(obj, [arg1, arg2...])
                    if(this.fun === Function.prototype.apply){
                        this.fun = this.target;
                        this.target = args[0] != null ? args[0] : AspectScript.globalObject;  //target is args[0]
                        args = args[1]; //args is args[1]
                        this.reflective = true;
                    }
                }
            };

            this.reinit(_kind, _target, _fun, _args, _context, _parent);

            this.__defineGetter__("args", function(){
                return Array.prototype.slice.call(args);
            });

            this.__defineGetter__("methods", function(){
                return (methods == null) ?
                       methods = findNamesOf(this.fun, this.target) :
                       Array.prototype.slice.call(methods); //clone methods array
            });

            this._proceed = function(){
                var actualArgs = (arguments.length == 0) ? args : arguments;
                return this.fun.apply(this.target, actualArgs);
            };

            this.clone = function(){
                var r = jpPool.get(this.kind, this.target, this.fun, args, this.context, this.parentAllLevel);
                r.copyFrom(this);
                return r;
            };

            this.toString = function(){
                return "[call: " + this.methods + "]"; //this.methods forces calculation of methods
            };
        }

        function ExecJoinPoint(_kind, _target, _args, _wrapperFun, _originalFun, _parent){
            JoinPoint.call(this);

            //private
            var args;
            var originalFun;
            var methods;

            this.reinit = function(_kind, _target, _args, _wrapperFun, _originalFun, _parent){
                this._reinit(_kind, _parent);
                this.target = _target;
                args = _args;
                this.fun = _wrapperFun;
                originalFun = _originalFun;
                methods = null;
            };

            this.reinit(_kind, _target, _args, _wrapperFun, _originalFun, _parent);

            this.__defineGetter__("args", function(){
                return Array.prototype.slice.call(args);
            });

            this.__defineGetter__("methods", function(){
                return (methods == null) ?
                       methods = findNamesOf(this.fun, this.target) :
                       Array.prototype.slice.call(methods); //clone methods array
            });

            this._proceed = function(){
                var actualArgs = (arguments.length == 0) ? args : arguments;
                return originalFun.apply(this.target, actualArgs);
            };

            this.clone = function(){
                var r = jpPool.get(this.kind, this.target, args, this.fun, originalFun, this.parentAllLevel);
                r.copyFrom(this);
                return r;
            };

            this.toString = function(){
                return "[exec: " + this.methods + "]"; //this.methods forces calculation of methods
            };
        }

        function PropertyReadJoinPoint(_kind, _target, _name, _parent){
            JoinPoint.call(this);

            this.reinit = function(_kind, _target, _name, _parent){
                this._reinit(_kind, _parent);
                this.target = _target;
                this.name = _name;
            };

            this.reinit(_kind, _target, _name, _parent);

            this._proceed = function(){
                return (arguments.length == 0) ? this.target[this.name] : arguments[0];
            };

            this.clone = function(){
                var r = jpPool.get(this.kind, this.target, this.name, this.parentAllLevel);
                r.copyFrom(this);
                return r;
            };

            this.toString = function(){
                return "[pr: " + this.name + "]";
            };
        }

        function PropertyWriteJoinPoint(_kind, _target, _name, _value, _parent){
            JoinPoint.call(this);

            this.reinit = function(_kind, _target, _name, _value, _parent){
                this._reinit(_kind, _parent);
                this.target = _target;
                this.name = _name;
                this.value = _value;
            };

            this.reinit(_kind, _target, _name, _value, _parent);

            this._proceed = function(){
                return this.target[this.name] = ((arguments.length == 0) ? this.value : arguments[0]);
            };

            this.clone = function(){
                var r = jpPool.get(this.kind, this.target, this.name, this.value, this.parentAllLevel);
                r.copyFrom(this);
                return r;
            };

            this.toString = function(){
                return "[pw: " + this.name + "]";
            };
        }

        function VarReadJoinPoint(_kind, _name, _value, _parent){
            JoinPoint.call(this);

            this.reinit = function(_kind, _name, _value, _parent){
                this._reinit(_kind, _parent);
                this.name = _name;
                this.value = _value;
            };

            this.reinit(_kind, _name, _value, _parent);

            this._proceed = function(){
                return (arguments.length == 0) ? this.value : arguments[0];
            };

            this.clone = function(){
                var r = jpPool.get(this.kind, this.name, this.value, this.parentAllLevel);
                r.copyFrom(this);
                return r;
            };

            this.toString = function(){
                return "[vr: " + this.name + "]";
            };
        }

        function VarWriteJoinPoint(_kind, _name, _value, _oldValue, _parent){
            JoinPoint.call(this);

            this.reinit = function(_kind, _name, _value, _oldValue, _parent){
                this._reinit(_kind, _parent);
                this.name = _name;
                this.value = _value;
                this.oldValue = _oldValue;
            };

            this.reinit(_kind, _name, _value, _oldValue, _parent);

            this._proceed = function(){
                return (arguments.length == 0) ? this.value : arguments[0];
            };

            this.clone = function(){
                var r = jpPool.get(this.kind, this.name, this.value, this.parentAllLevel);
                r.copyFrom(this);
                return r;
            };
        }

        //todo: how to change contextual info?
        //todo: what happens in the interaction of the pool and custom properties?
        function CustomJoinPoint(_kind, _name, _vars, _proceed, _parent){
            JoinPoint.call(this);

            //private
            var proceed;
            var vars;

            this.reinit = function(_kind, _name, _vars, _proceed, _parent){
                this._reinit(_kind, _parent);
                this.name = _name;

                vars = _vars;
                for(var i in _vars){
                    if(_vars.hasOwnProperty(i)){
                        this[i] = _vars[i];
                    }
                }

                proceed = _proceed;
            };

            this.reinit(_kind, _name, _vars, _proceed, _parent);

            this._proceed = function(){
                return proceed();
            };

            this.clone = function(){
                var r = jpPool.get(this.kind, this.name, vars, proceed, this.parentAllLevel);
                r.copyFrom(this);
                return r;
            };
        }

        function findNamesOf(fun, obj){
            var methods = [];
            for(var name in obj){
                //console.log("n: " + name);
                try{
                    if(obj[name] === fun){
                        methods.push(name);
                    }
                }
                catch(e){
                    //1.  domConfig
                    //console.log("conflictive name: " + name);
                }
            }
            return methods;
        }

        // Aspects
        //todo: the specified pc and adv are not properties of the resulting aspect (because of the wrapper funs)
        //todo:  => they do not have names
        function Aspect(aspKind, pointcut, advice, ctx){
            var aspect = this;

            var newAdvice = null;
            if(aspKind === AspectScript.BEFORE){
                newAdvice = function(jp, env){
                    var thunk = function(){
                        if(isWrapped(advice)){
                            advice(SILENCE_NEXT_EXEC_KEY);  //silence exec jp
                        }
                        advice.call(aspect, jp, env);
                        return jp.proceed();
                    };
                    return weave(jpPool.get(JP_EXEC, aspect, [jp, env], advice, thunk, currentJoinPoint));
                };
            }
            if(aspKind === AspectScript.AROUND){
                newAdvice = function(jp, env){
                    return advice.call(aspect, jp, env);
                };
            }
            if(aspKind === AspectScript.AFTER){
                newAdvice = function(jp, env){
                    var thunk = function(){
                        try{
                            return (jp.finalResult = jp.proceed());  //update finalResult for after advices
                        }
                        finally{
                            if(isWrapped(advice)){
                                advice(SILENCE_NEXT_EXEC_KEY);  //silence exec jp
                            }
                            advice.call(aspect, jp, env);
                        }
                    };

                    return weave(jpPool.get(JP_EXEC, aspect, [jp, env], advice, thunk, currentJoinPoint));
                };
            }

            this.advice = newAdvice;
            this.pointcut = function(jp, env){
                return pointcut.call(aspect, jp, env);
            };
            this.ctx = ctx;
        }

        var DEPLOYED_ASPECTS_IDS = 1;

        function wrapAspect(aspect, global, ss){
            return {
                id: DEPLOYED_ASPECTS_IDS++,
                pointcut: aspect.pointcut,
                advice: aspect.advice,
                isGlobal: global,
                c: ss[0],
                d: ss[1],
                f: ss[2],
                ctx: aspect.ctx
            };
        }

        //dd aspects
        //list of dd aspects (on objects/functions)
        var ddAspects = new AspectsMap();

        ddAspects.pushObjectsInContext(this.globalObject, function(){
        });

        function AspectsMap(){
            //value can be a single value or an array
            this.put = function(obj, value){
                if(typeof(obj.__ASPECTS__) == "undefined"){ //TODO
                    obj.__ASPECTS__ = [];
                }

                obj.__ASPECTS__ = obj.__ASPECTS__.concat(value);
            };

            this.remove = function(obj, value){
                if(typeof(obj.__ASPECTS__) == "undefined"){
                    return;
                }

                removeFromArray(obj.__ASPECTS__, value);
            };

            this.isEmpty = function(){
                return this.getAspectsForObjInContext().length == 0 &&
                       this.getAspectsForFunInContext().length == 0;
            };

            //object and function in context
            var contextObjs = [], contextFuns = [];

            this.pushObjectsInContext = function(obj, fun){
                contextObjs.push(obj);
                contextFuns.push(fun);
                tracer.pushCtx(this.getAspectsForObjInContext(), this.getAspectsForFunInContext());
            };

            this.popObjectsInContext = function(){
                tracer.popCtx(this.getAspectsForObjInContext(), this.getAspectsForFunInContext());
                contextObjs.pop();
                contextFuns.pop();
            };

            this.peekObjectsInContext = function(){
                return [contextObjs[contextObjs.length - 1], contextFuns[contextFuns.length - 1]];
            };

            this.getAspectsForObjInContext = function(){
                return ctxAspects(contextObjs);
            };

            this.getAspectsForFunInContext = function(){
                return ctxAspects(contextFuns);
            };

            function ctxAspects(objs){
                var currentObj = objs[objs.length - 1];

                if(typeof(currentObj.__ASPECTS__) == "undefined"){
                    return EMPTY_ARRAY;
                }

                return currentObj.__ASPECTS__;
            }
        }

        var aspectsWithSS = false; //todo: this should be a counter

        // ---- API for Aspects (Public)  ----

        this.aspect = function(aspKind, pc, adv, ctx){
            ctx = (typeof(ctx) !== "undefined") ? ctx : false;
            return new Aspect(aspKind, pc, adv, ctx);
        };

        this.before = function(pc, adv){
            return this.deploy(this.aspect(this.BEFORE, pc, adv, arguments[2]));
        };

        this.around = function(pc, adv){
            return this.deploy(this.aspect(this.AROUND, pc, adv, arguments[2]));
        };

        this.after = function(pc, adv){
            return this.deploy(this.aspect(this.AFTER, pc, adv, arguments[2]));
        };

        this.deploy = function(){
            tracer.deploy(currentMetaLevel + 1);
            var ss, aspect, fun;

            //global deployment
            if(arguments.length == 1){
                aspect = arguments[0];
                ss = [true, false, true];
                aspects.add(currentMetaLevel + 1, wrapAspect(aspect, true, ss));
                return aspect;
            }
            //dynamic deployment
            else{
                //ss parameter is optional
                var start;
                if(arguments.length == 2){
                    ss = [true, false, true];
                    start = 0;
                }
                else if(arguments.length == 3){
                    ss = arguments[0];
                    start = 1;
                    aspectsWithSS = true; //optimization
                }
                else{
                    throw "Wrong number of arguments";
                }

                aspect = arguments[start++];
                fun = arguments[start++];

                aspect = wrapAspect(aspect, false, ss);
                aspects.add(currentMetaLevel + 1, aspect); //deploy
                try{
                    return fun();  //in the case of dd aspects, the user does not get a reference to the aspect
                }
                finally{
                    //undeploy
                    aspects.remove(currentMetaLevel + 1, aspect);
                }
            }
        };

        //todo: update aspectsWithSS
        /*this.undeploy = function(aspect, obj){
         removeFromArray(aspect.metaLevels, currentMetaLevel + 1);

         //only aspect => global aspect
         if(arguments.length == 1){
         //static
         removeFromArray(aspects, aspect);
         //dd
         //ddAspects.removeAll(aspect); //??
         }
         //aspect and object
         else{
         ddAspects.remove(obj, aspect);
         }
         };*/

        this.deployOn = function(){
            var ss, aspect, obj;

            //ss parameter is optional
            var start = 0;
            if(arguments.length == 2){
                ss = [false, false, true];
                start = 0;
            }
            else if(arguments.length == 3){
                ss = arguments[0];
                start = 1;
                aspectsWithSS = true; //optimization
            }
            else{
                throw "Wrong number of arguments";
            }

            aspect = arguments[start++];
            obj = arguments[start++];

            ddAspects.put(obj, wrapAspect(aspect, false, ss));
            return aspect;
        };

        //---- Instrumentation (i13n) (PUBLIC) ----

        function isI13NTarget(target){
            return target === AspectScript.i13n || //direct call
                   i13nMethods.indexOf(target) >= 0; //reflective call (target is the function)
        }

        //objects whose methods invocations should not be intercepted
        this.registerSystemObject = function(obj){
            systemObjects.push(obj);
        };

        var systemObjects = [
            this,
            this.Pointcuts
        ];

        this.jpModel = {
            varRead: true,
            varWrite: true
        };

        var thunks = [
            function(ktor){
                return new ktor();
            },
            function(ktor, a){
                return new ktor(a);
            },
            function(ktor, a, b){
                return new ktor(a, b);
            },
            function(ktor, a, b, c){
                return new ktor(a, b, c);
            },
            function(ktor, a, b, c, d){
                return new ktor(a, b, c, d);
            },
            function(ktor, a, b, c, d, e){
                return new ktor(a, b, c, d, e);
            },
            function(ktor, a, b, c, d, e, f){
                return new ktor(a, b, c, d, e, f);
            },
            function(ktor, a, b, c, d, e, f, g){
                return new ktor(a, b, c, d, e, f, g);
            },
            function(ktor, a, b, c, d, e, f, g, h){
                return new ktor(a, b, c, d, e, f, g, h);
            },
            function(ktor, a, b, c, d, e, f, g, h, i){
                return new ktor(a, b, c, d, e, f, g, h, i);
            },
            function(ktor, a, b, c, d, e, f, g, h, i, j){
                return new ktor(a, b, c, d, e, f, g, h, i, j);
            }
        ];

        this.i13n = {
            wrap: function (thunk){
                var execKind = JP_EXEC;
                var wrapper = function(){

                    ///if the function is called with this special parameter, the exec join point is silenced next time
                    if(arguments[0] === SILENCE_NEXT_EXEC_KEY){
                        execKind = null;
                        return;
                    }

                    //the first time, we resolve the original fun
                    //(we use the thunk pattern to delay evaluation)
                    if(thunk.wrapper === undefined){
                        //we complete the creation of "this" in the initialization of functions
                        //which is the wrapper *and* the thunk, the proxy and the real object
                        try{
                            thunk = thunk.apply(null, arguments);
                        }
                        catch(e){
                            throw "thunk";
                        }
                        //update relationships between wrapper and thunk
                        thunk.wrapper = wrapper;

                        //now we are ready to create a new init jp for an *empty* initialization, as it is the case for functions
                        weave(jpPool.get(JP_INIT, wrapper, [], Function, returnTrueFunction, currentJoinPoint));

                        return wrapper;
                    }

                    var isConstructor = (this.constructor == wrapper); //todo: this instanceof wrapper?
                    if(isConstructor){
                        //don't return the value (which is "undefined" anyway)
                        weave(jpPool.get(JP_INIT, this, arguments, wrapper, thunk, currentJoinPoint));
                    }
                    else{
                        if(execKind == null){
                            execKind = JP_EXEC;  //back to exec
                            return thunk.apply(this, arguments);
                        }

                        return weave(jpPool.get(execKind, this, arguments, wrapper, thunk, currentJoinPoint));
                    }
                };

                //function declaration (creation)
                return weave(jpPool.get(JP_NEW, wrapper, Function, [], currentJoinPoint));
            },

            creation: function (thunk, constructor, args){
                if(thunk >= thunks.length){
                    throw "too few thunks: " + thunk;
                }
                thunk = thunks[thunk];

                if(constructor === Function){
                    var oldThunk = thunk;
                    thunk = function(){
                        return oldThunk.call(null, Function, args);
                    };
                    return this.wrap(thunk);
                }

                return weave(jpPool.get(JP_NEW, thunk, constructor, args, currentJoinPoint));
            },

            //literal object creation
            creation2: function (initializer, ctx, _arguments){
                var thunk = function(){
                    var obj = {};
                    //we have to use the context in which the object is being created
                    //(it has been changed with the creation of thunk)
                    var init = function(){
                        initializer.call(ctx, obj, _arguments);
                    };
                    weave(jpPool.get(JP_INIT, obj, [], Object, init, currentJoinPoint));  //ignore INIT result
                    return obj;
                };

                //return this.creation(thunk, Object, []);
                return weave(jpPool.get(JP_NEW, thunk, Object, [], currentJoinPoint));
            },

            // "{}" case, it is an optimization as we don't need initializer, ctx and arguments
            creation3: function (){
                var thunk = function(){
                    var obj = {};
                    weave(jpPool.get(JP_INIT, obj, [], Object, returnTrueFunction, currentJoinPoint));  //ignore INIT result
                    return obj;
                };

                return weave(jpPool.get(JP_NEW, thunk, Object, [], currentJoinPoint));
            },

            call: function (obj, fun, args, context){
                //just in case "call" is invoked like this: AspectScript.i13n.call()
                if((obj === null || typeof(obj) === "undefined") &&
                   (fun === null || typeof(fun) === "undefined")){
                    throw "cannot call AspectScript.i13n.call";
                }

                if(isI13NTarget(obj)){
                    throw "cannot call AspectScript.i13n methods directly";
                }

                var call = jpPool.get(JP_CALL, obj, fun, args, context, currentJoinPoint);
                if(systemObjects.indexOf(obj) >= 0){
                    try{
                        return call.proceed();
                    }
                    finally{
                        jpPool.release(call);
                    }
                }

                return weave(call);
            },

            //alternative case when obj IS specified (obj.m)
            call2: function(obj, funName, args, context){
                if(isI13NTarget(obj)){
                    throw "cannot call AspectScript.i13n methods directly";
                }

                if(typeof(obj) == "undefined"){
                    throw "\"" + funName + "\" function called on undefined";
                }

                if(obj == null){ //normally, this was a nullpointerexception
                    throw "obj is null";
                }

                var call = jpPool.get(JP_CALL, obj, this.propRead(obj, funName), args, context, currentJoinPoint);
                if(systemObjects.indexOf(obj) >= 0){
                    try{
                        return call.proceed();
                    }
                    finally{
                        jpPool.release(call);
                    }
                }

                return weave(call);
            },

            propRead: function(obj, name){
                return weave(jpPool.get(JP_PROP_READ, obj, name, currentJoinPoint));
            },

            propWrite: function (obj, propName, value){
                return weave(jpPool.get(JP_PROP_WRITE, obj, propName, value, currentJoinPoint));
            },

            varRead: function(name, value){
                if(!AspectScript.jpModel.varRead){
                    return value;
                }

                return weave(jpPool.get(JP_VAR_READ, name, value, currentJoinPoint));
            },

            varWrite: function(name, value, oldValue){
                if(!AspectScript.jpModel.varWrite){
                    return value;
                }

                return weave(jpPool.get(JP_VAR_WRITE, name, value, oldValue, currentJoinPoint));
            },

            propIncr : function(obj, prop, incr, postfix){
                var value = this.propWrite(obj, prop, this.propRead(obj, prop) + incr);
                return postfix ? value - incr : value;
            }
        };

        // Useful private vars for i13n
        var i13nMethods = [];
        for(var i in this.i13n){
            i13nMethods.push(this.i13n[i]);
        }

        var wrappedFunctionText = this.i13n.wrap(function(){
            return function(){
            };
        }).toString();

        function isWrapped(fun){
            return wrappedFunctionText === fun.toString();
        }

        //---- Reentrance (PRIVATE) ----

        function isReentering(jp, aspect){
            var ctx = aspect.ctx;

            if(ctx === true){
                return false;
            }

            var DEBUG = false;
            var ctxJP = ctx !== false ? ctx(jp) : false; //jp is JP in pc(JP), not a pc_exec in the case of nested pointcuts
            for(var i = 0; i < aspectsChain.length; ++i){
                //aspectsChain[i] = [aspect, jp]
                var aux = aspectsChain[i][0];
                var jp2 = aspectsChain[i][1];

                //first we ask whether the join points are at the same level
                if(jp2.metaLevel != jp.metaLevel){
                    continue; //todo: deployOn & levels!!
                }
                if(aspect === aux && (ctx === false || ctxJP === ctx(jp2))){
                    if(DEBUG){
                        print("ar");
                    }
                    return true;
                }
            }

            return false;
        }

        //---- API for Fluent Interface (PUBLIC) ----

        this.applyFluentInterface = function(fun){
            var pointcutsFluentInterface = {
                and: function(right){
                    var left = this;
                    return wrapPointcut(function(jp, env){
                        var newEnv = left(jp, env);
                        if(!newEnv){
                            return false;
                        }
                        return right(jp, newEnv);
                    });
                },
                or: function(right){
                    var left = this;
                    return wrapPointcut(function(jp, env){
                        var newEnv = left(jp, env);
                        if(!newEnv){
                            return right(jp, env);
                        }
                        return newEnv;
                    });
                },
                inCFlowOf: function(pc){
                    return this.and(AspectScript.Pointcuts.cflow(pc));
                }
            };

            for(var methodName in pointcutsFluentInterface){
                if(!fun[methodName]){
                    fun[methodName] = pointcutsFluentInterface[methodName];
                }
            }
            return fun;
        };

        function wrapPointcut(pcFun){
            var wrap = AspectScript.i13n.wrap(function(){
                return function (jp, env){
                    var newEnv = pcFun(jp, env);
                    if(newEnv && (newEnv instanceof Env)){
                        return newEnv;
                    }
                    return (newEnv) ? (env ? env : true) : false;
                };
            });

            return AspectScript.applyFluentInterface(wrap);
        }

        this.Pointcuts = {
            within:function(target){
                return wrapPointcut(function (jp){
                    //return jp.parent != null && this.cflow(pc)(jp.parent);
                    if((jp = jp.parent) == null){
                        return false;
                    }
                    do{
                        if(jp.target === target){
                            return true;
                        }

                        jp = jp.parent;
                    } while(jp != null && (jp.kind != JP_INIT || jp.kind != JP_EXEC));

                    return false;
                });
            },

            get: function(target, name){
                if(arguments.length == 1 && target === "*"){
                    return wrapPointcut(function (jp){
                        return jp.kind == JP_VAR_READ || jp.kind == JP_PROP_READ;
                    });
                }
                // by name only (var read)
                else if(arguments.length == 1){
                    name = target;
                    return wrapPointcut(function (jp){
                        return jp.kind == JP_VAR_READ && (jp.name === name || name === "*");
                    });
                }
                else{
                    // by name and target (prop read)
                    return wrapPointcut(function (jp){
                        return jp.kind == JP_PROP_READ
                                && (jp.target === target || target === "*")
                                && (jp.name === name || name === "*");
                    });
                }
            },

            //vars by name and props by (object, name)
            set: function(target, name){
                if(arguments.length == 1 && target === "*"){
                    return wrapPointcut(function (jp){
                        return jp.kind == JP_VAR_WRITE || jp.kind == JP_PROP_WRITE;
                    });
                }
                // by name only (var write)
                else if(arguments.length == 1){
                    name = target;
                    return wrapPointcut(function (jp){
                        return jp.kind == JP_VAR_WRITE && (jp.name === name || name === "*");
                    });
                }
                else{
                    // by name and target (prop write)
                    return wrapPointcut(function (jp){
                        return jp.kind == JP_PROP_WRITE
                                && (jp.target === target || target === "*")
                                && (jp.name === name || name === "*");
                    });
                }
            },

            creation: function(fun){
                //only by ref
                return wrapPointcut(function(jp){
                    return jp.kind == JP_NEW && jp.fun === fun;
                });
            },

            init: function(fun){
                if(arguments.length == 1 && fun === "*"){
                    return wrapPointcut(function (jp){
                        return jp.kind == JP_INIT;
                    });
                }
                //only by ref
                return wrapPointcut(function(jp){
                    return jp.kind == JP_INIT && jp.fun === fun;
                });
            },

            call: function(arg){
                // if arg is "*" (All calls)
                if(typeof arg == "string" && arg === "*"){
                    return function(jp){
                        return jp.kind == JP_CALL;
                    };
                }
                //by name
                if(typeof arg == "string"){
                    return wrapPointcut(function (jp){
                        return jp.kind == JP_CALL && jp.methods.indexOf(arg) >= 0;
                    });
                }
                //by ref
                else{
                    return wrapPointcut(function(jp){
                        return jp.kind == JP_CALL && jp.fun === arg;
                    });
                }
            },

            exec: function(arg){
                // if arg is "*" (All execs)
                if(typeof arg == "string" && arg === "*"){
                    return function(jp){
                        return jp.kind == JP_EXEC;
                    };
                }
                //by name
                if(typeof arg == "string"){
                    return wrapPointcut(function (jp){
                        return jp.kind == JP_EXEC && jp.methods.indexOf(arg) >= 0;
                    });
                }
                //by ref
                else{
                    return wrapPointcut(function(jp){
                        return jp.kind == JP_EXEC && jp.fun === arg;
                    });
                }
            },

            target: function (obj){
                return wrapPointcut(function(jp){
                    if(obj === "*" && jp.target !== undefined){
                        return true;
                    }
                    return jp.target === obj;
                });
            },

            cflow: function (pc){
                return wrapPointcut(function (jp, env){
                    //return pc(jp) || this.cflowbelow(pc)(jp);
                    while(jp != null){
                        var newEnv = pc(jp, env);
                        if(newEnv){
                            return newEnv;
                        }
                        jp = jp.parent;
                    }
                    return false;
                });
            },

            cflowbelow: function(pc){
                return wrapPointcut(function (jp, env){
                    //return jp.parent != null && this.cflow(pc)(jp.parent);
                    jp = jp.parent;
                    while(jp != null){
                        var newEnv = pc(jp, env);
                        if(newEnv){
                            return newEnv;
                        }
                        jp = jp.parent;
                    }
                    return false;
                });
            },

            event: function (name){
                return wrapPointcut(function(jp){
                    return jp.kind == JP_CUSTOM &&
                           (jp.name === undefined || jp.name === name); //all custom or the name matches
                });
            },

            not: function(pc){
                return wrapPointcut(function (jp, env){
                    return !pc(jp, env);
                });
            },

            noBR: function(pc, ctx){
                return wrapPointcut(function(jp, env){
                    var pcBR = wrapPointcut(pc);
                    if(ctx !== undefined){
                        var ctxJP = ctx(jp);
                        pcBR = pcBR.and(function(jp){
                            return ctxJP === ctx(jp);
                        });
                    }
                    if(!AspectScript.Pointcuts.cflowbelow(pcBR)(jp, env)){
                        return pc(jp, env);
                    }
                });
            }
        };

        // ---- API CUSTOM JOIN POINT (Public)

        this.event = function(name, args, proceed){
            return weave(jpPool.get(JP_CUSTOM, name, args, proceed, currentJoinPoint));
        };

        //---- API UP & DOWN (Public)

        this.down = function(obj, exp){
            --currentMetaLevel;
            try{
                if(arguments.length == 1){  //obj is the function
                    return obj();
                }
                else{
                    return exp.apply(obj);
                }
            }
            finally{
                ++currentMetaLevel;
            }
        };

        this.up = function (obj, exp){
            ++currentMetaLevel;
            try{
                if(arguments.length == 1){  //obj is the function
                    return obj();
                }
                else{
                    return exp.apply(obj);
                }
            }
            finally{
                --currentMetaLevel;
            }
        };
    }

    //singleton
    try{
        weavingEnabled = false;
        return new Constructor();
    }
    finally{
        weavingEnabled = true;
    }
})();
